# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import warnings

import cirq
from qsimcirq import qsim
from typing import Dict, Union


# List of parameter names that appear in valid Cirq protos.
GATE_PARAMS = [
  'exponent', 'phase_exponent', 'global_shift',
  'x_exponent', 'z_exponent', 'axis_phase_exponent',
  'phi', 'theta'
]


def _cirq_gate_kind(gate: cirq.ops.Gate):
  if isinstance(gate, cirq.ops.ControlledGate):
    return _cirq_gate_kind(gate.sub_gate)
  if isinstance(gate, cirq.ops.identity.IdentityGate):
    if gate.num_qubits() == 1:
      return qsim.kI1
    if gate.num_qubits() == 2:
      return qsim.kI2
    if gate.num_qubits() <= 6:
      return qsim.kI
    raise NotImplementedError(
      f'Received identity on {gate.num_qubits()} qubits; '
      + 'only up to 6-qubit gates are supported.')
  if isinstance(gate, cirq.ops.XPowGate):
    # cirq.rx also uses this path.
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kX
    return qsim.kXPowGate
  if isinstance(gate, cirq.ops.YPowGate):
    # cirq.ry also uses this path.
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kY
    return qsim.kYPowGate
  if isinstance(gate, cirq.ops.ZPowGate):
    # cirq.rz also uses this path.
    if gate.global_shift == 0:
      if gate.exponent == 1:
        return qsim.kZ
      if gate.exponent == 0.5:
        return qsim.kS
      if gate.exponent == 0.25:
        return qsim.kT
    return qsim.kZPowGate
  if isinstance(gate, cirq.ops.HPowGate):
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kH
    return qsim.kHPowGate
  if isinstance(gate, cirq.ops.CZPowGate):
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kCZ
    return qsim.kCZPowGate
  if isinstance(gate, cirq.ops.CXPowGate):
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kCX
    return qsim.kCXPowGate
  if isinstance(gate, cirq.ops.PhasedXPowGate):
    return qsim.kPhasedXPowGate
  if isinstance(gate, cirq.ops.PhasedXZGate):
    return qsim.kPhasedXZGate
  if isinstance(gate, cirq.ops.XXPowGate):
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kXX
    return qsim.kXXPowGate
  if isinstance(gate, cirq.ops.YYPowGate):
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kYY
    return qsim.kYYPowGate
  if isinstance(gate, cirq.ops.ZZPowGate):
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kZZ
    return qsim.kZZPowGate
  if isinstance(gate, cirq.ops.SwapPowGate):
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kSWAP
    return qsim.kSwapPowGate
  if isinstance(gate, cirq.ops.ISwapPowGate):
    # cirq.riswap also uses this path.
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kISWAP
    return qsim.kISwapPowGate
  if isinstance(gate, cirq.ops.PhasedISwapPowGate):
    # cirq.givens also uses this path.
    return qsim.kPhasedISwapPowGate
  if isinstance(gate, cirq.ops.FSimGate):
    return qsim.kFSimGate
  if isinstance(gate, cirq.ops.TwoQubitDiagonalGate):
    return qsim.kTwoQubitDiagonalGate
  if isinstance(gate, cirq.ops.ThreeQubitDiagonalGate):
    return qsim.kThreeQubitDiagonalGate
  if isinstance(gate, cirq.ops.CCZPowGate):
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kCCZ
    return qsim.kCCZPowGate
  if isinstance(gate, cirq.ops.CCXPowGate):
    if gate.exponent == 1 and gate.global_shift == 0:
      return qsim.kCCX
    return qsim.kCCXPowGate
  if isinstance(gate, cirq.ops.CSwapGate):
    return qsim.kCSwapGate
  if isinstance(gate, cirq.ops.MatrixGate):
    if gate.num_qubits() <= 6:
      return qsim.kMatrixGate
    raise NotImplementedError(
      f'Received matrix on {gate.num_qubits()} qubits; '
      + 'only up to 6-qubit gates are supported.')
  if isinstance(gate, cirq.ops.MeasurementGate):
    # needed to inherit SimulatesSamples in sims
    return qsim.kMeasurement
  # Unrecognized gates will be decomposed.
  return None


def _control_details(gate: cirq.ops.ControlledGate, qubits):
  control_qubits = []
  control_values = []
  # TODO: support qudit control
  for i, cvs in enumerate(gate.control_values):
    if 0 in cvs and 1 in cvs:
      # This qubit does not affect control.
      continue
    elif 0 not in cvs and 1 not in cvs:
      # This gate will never trigger.
      warnings.warn(f'Gate has no valid control value: {gate}', RuntimeWarning)
      return (None, None)
    # Either 0 or 1 is in cvs, but not both.
    control_qubits.append(qubits[i])
    if 0 in cvs:
      control_values.append(0)
    elif 1 in cvs:
      control_values.append(1)

  return (control_qubits, control_values)


def add_op_to_opstring(
    qsim_op: cirq.GateOperation,
    qubit_to_index_dict: Dict[cirq.Qid, int],
    opstring: qsim.OpString,
):
  """Adds an operation to an opstring (observable).
  
  Raises:
    ValueError if qsim_op is not a single-qubit Pauli (I, X, Y, or Z).
  """
  qsim_gate = qsim_op.gate
  gate_kind = _cirq_gate_kind(qsim_gate)
  if gate_kind not in {qsim.kX, qsim.kY, qsim.kZ, qsim.kI1}:
    raise ValueError(f'OpString should only have Paulis; got {gate_kind}')
  if len(qsim_op.qubits) != 1:
    raise ValueError(
      f'OpString ops should have 1 qubit; got {len(qsim_op.qubits)}')

  is_controlled = isinstance(qsim_gate, cirq.ops.ControlledGate)
  if is_controlled:
    raise ValueError(f'OpString ops should not be controlled.')

  qubits = [qubit_to_index_dict[q] for q in qsim_op.qubits]
  qsim.add_gate_to_opstring(gate_kind, qubits, opstring)


def add_op_to_circuit(
    qsim_op: cirq.GateOperation,
    time: int,
    qubit_to_index_dict: Dict[cirq.Qid, int],
    circuit: Union[qsim.Circuit, qsim.NoisyCircuit]
):
  """Adds an operation to a noisy or noiseless circuit."""
  qsim_gate = qsim_op.gate
  gate_kind = _cirq_gate_kind(qsim_gate)
  qubits = [qubit_to_index_dict[q] for q in qsim_op.qubits]

  qsim_qubits = qubits
  is_controlled = isinstance(qsim_gate, cirq.ops.ControlledGate)
  if is_controlled:
    control_qubits, control_values = _control_details(qsim_gate, qubits)
    if control_qubits is None:
      # This gate has no valid control, and will be omitted.
      return

    if qsim_gate.num_qubits() > 4:
      raise NotImplementedError(
      f'Received control gate on {gate.num_qubits()} target qubits; '
      + 'only up to 4-qubit gates are supported.')

    qsim_qubits = qubits[qsim_gate.num_controls():]
    qsim_gate = qsim_gate.sub_gate

  if (gate_kind == qsim.kTwoQubitDiagonalGate or
      gate_kind == qsim.kThreeQubitDiagonalGate):
    if isinstance(circuit, qsim.Circuit):
      qsim.add_diagonal_gate(
        time, qsim_qubits, qsim_gate._diag_angles_radians, circuit)
    else:
      qsim.add_diagonal_gate_channel(
        time, qsim_qubits, qsim_gate._diag_angles_radians, circuit)
  elif gate_kind == qsim.kMatrixGate:
    m = [
      val for i in list(cirq.unitary(qsim_gate).flat)
      for val in [i.real, i.imag]
    ]
    if isinstance(circuit, qsim.Circuit):
      qsim.add_matrix_gate(time, qsim_qubits, m, circuit)
    else:
      qsim.add_matrix_gate_channel(time, qsim_qubits, m, circuit)
  else:
    params = {}
    for p, val in vars(qsim_gate).items():
      key = p.strip('_')
      if key not in GATE_PARAMS:
        continue
      if isinstance(val, (int, float, np.integer, np.floating)):
        params[key] = val
      else:
        raise ValueError('Parameters must be numeric.')
    if isinstance(circuit, qsim.Circuit):
      qsim.add_gate(gate_kind, time, qsim_qubits, params, circuit)
    else:
      qsim.add_gate_channel(gate_kind, time, qsim_qubits, params, circuit)

  if is_controlled:
    if isinstance(circuit, qsim.Circuit):
      qsim.control_last_gate(control_qubits, control_values, circuit)
    else:
      qsim.control_last_gate_channel(control_qubits, control_values, circuit)


class QSimCircuit(cirq.Circuit):

  def __init__(self,
               cirq_circuit: cirq.Circuit,
               device: cirq.devices = cirq.devices.UNCONSTRAINED_DEVICE,
               allow_decomposition: bool = False):

    if allow_decomposition:
      super().__init__([], device=device)
      for moment in cirq_circuit:
        for op in moment:
          # This should call decompose on the gates
          self.append(op)
    else:
      super().__init__(cirq_circuit, device=device)

  def __eq__(self, other):
    if not isinstance(other, QSimCircuit):
      return False
    # equality is tested, for the moment, for cirq.Circuit
    return super().__eq__(other)

  def _resolve_parameters_(self, param_resolver: cirq.study.ParamResolver):
    return QSimCircuit(
      cirq.resolve_parameters(super(), param_resolver), device=self.device)

  def translate_cirq_to_qsim(
      self,
      qubit_order: cirq.ops.QubitOrderOrList = cirq.ops.QubitOrder.DEFAULT
  ) -> qsim.Circuit:
    """
        Translates this Cirq circuit to the qsim representation.
        :qubit_order: Ordering of qubits
        :return: a C++ qsim Circuit object
        """

    qsim_circuit = qsim.Circuit()
    qubits = self.all_qubits()
    qsim_circuit.num_qubits = len(qubits)
    ordered_qubits = cirq.ops.QubitOrder.as_qubit_order(qubit_order).order_for(
      qubits)

    # qsim numbers qubits in reverse order from cirq
    ordered_qubits = list(reversed(ordered_qubits))

    def has_qsim_kind(op: cirq.ops.GateOperation):
      return _cirq_gate_kind(op.gate) != None

    def to_matrix(op: cirq.ops.GateOperation):
      mat = cirq.protocols.unitary(op.gate, None)
      if mat is None:
          return NotImplemented
      
      return cirq.ops.MatrixGate(mat).on(*op.qubits)

    qubit_to_index_dict = {q: i for i, q in enumerate(ordered_qubits)}
    time_offset = 0
    for moment in self:
      ops_by_gate = [
        cirq.decompose(op, fallback_decomposer=to_matrix, keep=has_qsim_kind)
        for op in moment
      ]
      moment_length = max(len(gate_ops) for gate_ops in ops_by_gate)

      # Gates must be added in time order.
      for gi in range(moment_length):
        for gate_ops in ops_by_gate:
          if gi >= len(gate_ops):
            continue
          qsim_op = gate_ops[gi]
          time = time_offset + gi
          gate_kind = _cirq_gate_kind(qsim_op.gate)
          add_op_to_circuit(qsim_op, time, qubit_to_index_dict, qsim_circuit)
      time_offset += moment_length

    return qsim_circuit


  def translate_cirq_to_qtrajectory(
      self,
      qubit_order: cirq.ops.QubitOrderOrList = cirq.ops.QubitOrder.DEFAULT
  ) -> qsim.NoisyCircuit:
    """
        Translates this noisy Cirq circuit to the qsim representation.
        :qubit_order: Ordering of qubits
        :return: a C++ qsim NoisyCircuit object
        """
    qsim_ncircuit = qsim.NoisyCircuit()
    ordered_qubits = cirq.ops.QubitOrder.as_qubit_order(qubit_order).order_for(
        self.all_qubits())

    # qsim numbers qubits in reverse order from cirq
    ordered_qubits = list(reversed(ordered_qubits))

    def has_qsim_kind(op: cirq.ops.GateOperation):
      return _cirq_gate_kind(op.gate) != None

    def to_matrix(op: cirq.ops.GateOperation):
      mat = cirq.unitary(op.gate, None)
      if mat is None:
          return NotImplemented
      
      return cirq.ops.MatrixGate(mat).on(*op.qubits)

    qubit_to_index_dict = {q: i for i, q in enumerate(ordered_qubits)}
    time_offset = 0
    for moment in self:
      moment_length = 0
      ops_by_gate = []
      ops_by_mix = []
      ops_by_channel = []
      # Capture ops of each type in the appropriate list.
      for qsim_op in moment:
        if cirq.has_unitary(qsim_op) or cirq.is_measurement(qsim_op):
          oplist = cirq.decompose(
            qsim_op, fallback_decomposer=to_matrix, keep=has_qsim_kind)
          ops_by_gate.append(oplist)
          moment_length = max(moment_length, len(oplist))
          pass
        elif cirq.has_mixture(qsim_op):
          ops_by_mix.append(qsim_op)
          moment_length = max(moment_length, 1)
          pass
        elif cirq.has_channel(qsim_op):
          ops_by_channel.append(qsim_op)
          moment_length = max(moment_length, 1)
          pass
        else:
          raise ValueError(f'Encountered unparseable op: {qsim_op}')

      # Gates must be added in time order.
      for gi in range(moment_length):
        # Handle gate output.
        for gate_ops in ops_by_gate:
          if gi >= len(gate_ops):
            continue
          qsim_op = gate_ops[gi]
          time = time_offset + gi
          gate_kind = _cirq_gate_kind(qsim_op.gate)
          add_op_to_circuit(qsim_op, time, qubit_to_index_dict, qsim_ncircuit)
        # Handle mixture output.
        for mixture in ops_by_mix:
          mixdata = []
          for prob, mat in cirq.mixture(mixture):
            square_mat = np.reshape(mat, (int(np.sqrt(mat.size)), -1))
            unitary = cirq.is_unitary(square_mat)
            # Package matrix into a qsim-friendly format.
            mat = np.reshape(mat, (-1,)).astype(np.complex64, copy=False)
            mixdata.append((prob, mat.view(np.float32), unitary))
          qubits = [qubit_to_index_dict[q] for q in mixture.qubits]
          qsim.add_channel(time_offset, qubits, mixdata, qsim_ncircuit)
        # Handle channel output.
        for channel in ops_by_channel:
          chdata = []
          for i, mat in enumerate(cirq.channel(channel)):
            square_mat = np.reshape(mat, (int(np.sqrt(mat.size)), -1))
            unitary = cirq.is_unitary(square_mat)
            singular_vals = np.linalg.svd(square_mat)[1]
            lower_bound_prob = min(singular_vals) ** 2
            # Package matrix into a qsim-friendly format.
            mat = np.reshape(mat, (-1,)).astype(np.complex64, copy=False)
            chdata.append((lower_bound_prob, mat.view(np.float32), unitary))
          qubits = [qubit_to_index_dict[q] for q in channel.qubits]
          qsim.add_channel(time_offset, qubits, chdata, qsim_ncircuit)
      time_offset += moment_length

    return qsim_ncircuit



# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from cirq import (
  circuits,
  linalg,
  ops,
  protocols,
  sim,
  study,
  value,
  SimulatesAmplitudes,
  SimulatesFinalState,
  SimulatesSamples,
)

import numpy as np

from qsimcirq import qsim
import qsimcirq.qsim_circuit as qsimc


class QSimSimulatorState(sim.StateVectorSimulatorState):

    def __init__(self,
                 qsim_data: np.ndarray,
                 qubit_map: Dict[ops.Qid, int]):
      state_vector = qsim_data.view(np.complex64)
      super().__init__(state_vector=state_vector, qubit_map=qubit_map)


class QSimSimulatorTrialResult(sim.StateVectorTrialResult):

    def __init__(self,
                 params: study.ParamResolver,
                 measurements: Dict[str, np.ndarray],
                 final_simulator_state: QSimSimulatorState):
      super().__init__(params=params,
                       measurements=measurements,
                       final_simulator_state=final_simulator_state)


# This should probably live in Cirq...
# TODO: update to support CircuitOperations.
def _needs_trajectories(circuit: circuits.Circuit) -> bool:
  """Checks if the circuit requires trajectory simulation."""
  for op in circuit.all_operations():
    test_op = (
      op if not protocols.is_parameterized(op)
      else protocols.resolve_parameters(
          op, {param: 1 for param in protocols.parameter_names(op)}
        )
    )
    if not (protocols.has_unitary(test_op) or protocols.is_measurement(test_op)):
      return True
  return False


class QSimSimulator(SimulatesSamples, SimulatesAmplitudes, SimulatesFinalState):

  def __init__(self, qsim_options: dict = {},
               seed: value.RANDOM_STATE_OR_SEED_LIKE = None):
    """Creates a new QSimSimulator using the given options and seed.
    
    Args:
        qsim_options: A map of circuit options for the simulator. These will be
            applied to all circuits run using this simulator. Accepted keys and
            their behavior are as follows:
                - 'f': int (> 0). Maximum size of fused gates. Default: 2.
                - 't': int (> 0). Number of threads to run on. Default: 1.
                - 'v': int (>= 0). Log verbosity. Default: 0.
            See qsim/docs/usage.md for more details on these options.
        seed: A random state or seed object, as defined in cirq.value.
    
    Raises:
        ValueError if internal keys 'c' or 'i' are included in 'qsim_options'.
    """
    if any(k in qsim_options for k in ('c', 'i', 'n')):
      raise ValueError(
          'Keys {"c", "i", "n"} are reserved for internal use and cannot be '
          'used in QSimCircuit instantiation.'
      )
    self._prng = value.parse_random_state(seed)
    self.qsim_options = {'t': 1, 'f': 2, 'v': 0}
    self.qsim_options.update(qsim_options)

  def get_seed(self):
    # Limit seed size to 32-bit integer for C++ conversion.
    return self._prng.randint(2 ** 31 - 1)

  def _run(
        self,
        circuit: circuits.Circuit,
        param_resolver: study.ParamResolver,
        repetitions: int
  ) -> Dict[str, np.ndarray]:
    """Run a simulation, mimicking quantum hardware.

    Args:
        program: The circuit to simulate.
        param_resolver: Parameters to run with the program.
        repetitions: Number of times to repeat the run.

    Returns:
        A dictionary from measurement gate key to measurement
        results.
    """
    param_resolver = param_resolver or study.ParamResolver({})
    solved_circuit = protocols.resolve_parameters(circuit, param_resolver)

    return self._sample_measure_results(solved_circuit, repetitions)

  def _sample_measure_results(
    self,
    program: circuits.Circuit,
    repetitions: int = 1,
  ) -> Dict[str, np.ndarray]:
    """Samples from measurement gates in the circuit.

    Note that this will execute the circuit 'repetitions' times.

    Args:
        program: The circuit to sample from.
        repetitions: The number of samples to take.

    Returns:
        A dictionary from measurement gate key to measurement
        results. Measurement results are stored in a 2-dimensional
        numpy array, the first dimension corresponding to the repetition
        and the second to the actual boolean measurement results (ordered
        by the qubits being measured.)

    Raises:
        ValueError: If there are multiple MeasurementGates with the same key,
            or if repetitions is negative.
    """
    if not isinstance(program, qsimc.QSimCircuit):
      program = qsimc.QSimCircuit(program, device=program.device)

    # Compute indices of measured qubits
    ordered_qubits = ops.QubitOrder.DEFAULT.order_for(program.all_qubits())
    num_qubits = len(ordered_qubits)

    qubit_map = {
      qubit: index for index, qubit in enumerate(ordered_qubits)
    }

    # Computes
    # - the list of qubits to be measured
    # - the start (inclusive) and end (exclusive) indices of each measurement
    # - a mapping from measurement key to measurement gate
    measurement_ops = [
      op for _, op, _ in program.findall_operations_with_gate_type(ops.MeasurementGate)
    ]
    measured_qubits = []  # type: List[ops.Qid]
    bounds = {}  # type: Dict[str, Tuple]
    meas_ops = {}  # type: Dict[str, cirq.GateOperation]
    current_index = 0
    for op in measurement_ops:
      gate = op.gate
      key = protocols.measurement_key(gate)
      meas_ops[key] = op
      if key in bounds:
        raise ValueError("Duplicate MeasurementGate with key {}".format(key))
      bounds[key] = (current_index, current_index + len(op.qubits))
      measured_qubits.extend(op.qubits)
      current_index += len(op.qubits)

    # Set qsim options
    options = {}
    options.update(self.qsim_options)

    results = {}
    for key, bound in bounds.items():
      results[key] = np.ndarray(shape=(repetitions, bound[1]-bound[0]),
                                dtype=int)


    noisy = _needs_trajectories(program)
    if noisy:
      translator_fn_name = 'translate_cirq_to_qtrajectory'
      sampler_fn = qsim.qtrajectory_sample
    else:
      translator_fn_name = 'translate_cirq_to_qsim'
      sampler_fn = qsim.qsim_sample

    if not noisy and program.are_all_measurements_terminal() and repetitions > 1:
      print('Provided circuit has no intermediate measurements. ' +
            'Sampling repeatedly from final state vector.')
      # Measurements must be replaced with identity gates to sample properly.
      # Simply removing them may omit qubits from the circuit.
      for i in range(len(program.moments)):
        program.moments[i] = ops.Moment(
          op if not isinstance(op.gate, ops.MeasurementGate)
          else [ops.IdentityGate(1).on(q) for q in op.qubits]
          for op in program.moments[i]
        )
      options['c'] = program.translate_cirq_to_qsim(ops.QubitOrder.DEFAULT)
      options['s'] = self.get_seed()
      options['n'] = num_qubits
      final_state = qsim.qsim_simulate_fullstate(options, 0)
      full_results = sim.sample_state_vector(
        final_state.view(np.complex64), range(num_qubits),
        repetitions=repetitions, seed=self._prng)

      for i in range(repetitions):
        for key, op in meas_ops.items():
          meas_indices = [qubit_map[qubit] for qubit in op.qubits]
          for j, q in enumerate(meas_indices):
            results[key][i][j] = full_results[i][q]
    else:
      translator_fn = getattr(program, translator_fn_name)
      options['c'] = translator_fn(ops.QubitOrder.DEFAULT)
      for i in range(repetitions):
        options['s'] = self.get_seed()
        options['n'] = num_qubits
        measurements = sampler_fn(options)
        for key, bound in bounds.items():
          for j in range(bound[1]-bound[0]):
            results[key][i][j] = int(measurements[bound[0]+j])

    return results


  def compute_amplitudes_sweep(
      self,
      program: circuits.Circuit,
      bitstrings: Sequence[int],
      params: study.Sweepable,
      qubit_order: ops.QubitOrderOrList = ops.QubitOrder.DEFAULT,
  ) -> Sequence[Sequence[complex]]:
    """Computes the desired amplitudes using qsim.

      The initial state is assumed to be the all zeros state.

      Args:
          program: The circuit to simulate.
          bitstrings: The bitstrings whose amplitudes are desired, input as an
            string array where each string is formed from measured qubit values
            according to `qubit_order` from most to least significant qubit,
            i.e. in big-endian ordering.
          param_resolver: Parameters to run with the program.
          qubit_order: Determines the canonical ordering of the qubits. This is
            often used in specifying the initial state, i.e. the ordering of the
            computational basis states.

      Returns:
          List of amplitudes.
      """
    if not isinstance(program, qsimc.QSimCircuit):
      program = qsimc.QSimCircuit(program, device=program.device)

    num_qubits = len(program.all_qubits())
    # qsim numbers qubits in reverse order from cirq
    bitstrings = [format(bitstring, 'b').zfill(num_qubits)[::-1]
                  for bitstring in bitstrings]

    options = {'i': '\n'.join(bitstrings)}
    options.update(self.qsim_options)

    param_resolvers = study.to_resolvers(params)

    trials_results = []
    if _needs_trajectories(program):
      translator_fn_name = 'translate_cirq_to_qtrajectory'
      simulator_fn = qsim.qtrajectory_simulate
    else:
      translator_fn_name = 'translate_cirq_to_qsim'
      simulator_fn = qsim.qsim_simulate

    for prs in param_resolvers:
      solved_circuit = protocols.resolve_parameters(program, prs)
      translator_fn = getattr(solved_circuit, translator_fn_name)
      options['c'] = translator_fn(qubit_order)
      options['s'] = self.get_seed()
      options['n'] = num_qubits
      amplitudes = simulator_fn(options)
      trials_results.append(amplitudes)

    return trials_results

  def simulate_sweep(
      self,
      program: circuits.Circuit,
      params: study.Sweepable,
      qubit_order: ops.QubitOrderOrList = ops.QubitOrder.DEFAULT,
      initial_state: Optional[Union[int, np.ndarray]] = None,
  ) -> List['SimulationTrialResult']:
    """Simulates the supplied Circuit.

      This method returns a result which allows access to the entire
      wave function. In contrast to simulate, this allows for sweeping
      over different parameter values.

      Args:
          program: The circuit to simulate.
          params: Parameters to run with the program.
          qubit_order: Determines the canonical ordering of the qubits. This is
            often used in specifying the initial state, i.e. the ordering of the
            computational basis states.
          initial_state: The initial state for the simulation. This can either
            be an integer representing a pure state (e.g. 11010) or a numpy
            array containing the full state vector. If none is provided, this
            is assumed to be the all-zeros state.

      Returns:
          List of SimulationTrialResults for this run, one for each
          possible parameter resolver.

      Raises:
          TypeError: if an invalid initial_state is provided.
      """
    if initial_state is None:
      initial_state = 0
    if not isinstance(initial_state, (int, np.ndarray)):
      raise TypeError('initial_state must be an int or state vector.')
    if not isinstance(program, qsimc.QSimCircuit):
      program = qsimc.QSimCircuit(program, device=program.device)

    options = {}
    options.update(self.qsim_options)

    param_resolvers = study.to_resolvers(params)
    qubits = program.all_qubits()
    num_qubits = len(qubits)
    if isinstance(initial_state, np.ndarray):
      if initial_state.dtype != np.complex64:
        raise TypeError(f'initial_state vector must have dtype np.complex64.')
      input_vector = initial_state.view(np.float32)
      if len(input_vector) != 2**num_qubits * 2:
        raise ValueError(f'initial_state vector size must match number of qubits.'
          f'Expected: {2**num_qubits * 2} Received: {len(input_vector)}')

    trials_results = []
    if _needs_trajectories(program):
      translator_fn_name = 'translate_cirq_to_qtrajectory'
      fullstate_simulator_fn = qsim.qtrajectory_simulate_fullstate
    else:
      translator_fn_name = 'translate_cirq_to_qsim'
      fullstate_simulator_fn = qsim.qsim_simulate_fullstate

    for prs in param_resolvers:
      solved_circuit = protocols.resolve_parameters(program, prs)
      translator_fn = getattr(solved_circuit, translator_fn_name)
      options['c'] = translator_fn(qubit_order)
      options['s'] = self.get_seed()
      options['n'] = num_qubits
      ordered_qubits = ops.QubitOrder.as_qubit_order(qubit_order).order_for(
        qubits)
      # qsim numbers qubits in reverse order from cirq
      ordered_qubits = list(reversed(ordered_qubits))

      qubit_map = {
        qubit: index for index, qubit in enumerate(ordered_qubits)
      }

      if isinstance(initial_state, int):
        qsim_state = fullstate_simulator_fn(options, initial_state)
      elif isinstance(initial_state, np.ndarray):
        qsim_state = fullstate_simulator_fn(options, input_vector)
      assert qsim_state.dtype == np.float32
      assert qsim_state.ndim == 1
      final_state = QSimSimulatorState(qsim_state, qubit_map)
      # create result for this parameter
      # TODO: We need to support measurements.
      result = QSimSimulatorTrialResult(params=prs,
                                        measurements={},
                                        final_simulator_state=final_state)
      trials_results.append(result)

    return trials_results

  def simulate_expectation_values_sweep(
    self,
    program: 'cirq.Circuit',
    observables: Union['cirq.PauliSumLike', List['cirq.PauliSumLike']],
    params: 'study.Sweepable',
    qubit_order: ops.QubitOrderOrList = ops.QubitOrder.DEFAULT,
    initial_state: Any = None,
    permit_terminal_measurements: bool = False,
  ) -> List[List[float]]:
    """Simulates the supplied circuit and calculates exact expectation
    values for the given observables on its final state.

    This method has no perfect analogy in hardware. Instead compare with
    Sampler.sample_expectation_values, which calculates estimated
    expectation values by sampling multiple times.

    Args:
        program: The circuit to simulate.
        observables: An observable or list of observables.
        param_resolver: Parameters to run with the program.
        qubit_order: Determines the canonical ordering of the qubits. This
            is often used in specifying the initial state, i.e. the
            ordering of the computational basis states.
        initial_state: The initial state for the simulation. The form of
            this state depends on the simulation implementation. See
            documentation of the implementing class for details.
        permit_terminal_measurements: If the provided circuit ends with
            measurement(s), this method will generate an error unless this
            is set to True. This is meant to prevent measurements from
            ruining expectation value calculations.

    Returns:
        A list of expectation values, with the value at index `n`
        corresponding to `observables[n]` from the input.

    Raises:
        ValueError if 'program' has terminal measurement(s) and
        'permit_terminal_measurements' is False. (Note: We cannot test this
        until Cirq's `are_any_measurements_terminal` is released.)
    """
    # TODO: replace with commented check when Cirq v0.10 is released.
    if not permit_terminal_measurements:
      raise ValueError(
        'Automatic terminal measurement checking is not supported in qsim. '
        'Please check that your circuit has no terminal measurements, then '
        'set permit_terminal_measurements=True to bypass this error.'
      )
    # if not permit_terminal_measurements and program.are_any_measurements_terminal():
    #   raise ValueError(
    #     'Provided circuit has terminal measurements, which may '
    #     'skew expectation values. If this is intentional, set '
    #     'permit_terminal_measurements=True.'
    #   )
    if not isinstance(observables, List):
      observables = [observables]
    psumlist = [ops.PauliSum.wrap(pslike) for pslike in observables]

    ordered_qubits = ops.QubitOrder.as_qubit_order(qubit_order).order_for(
      program.all_qubits())
    ordered_qubits = list(reversed(ordered_qubits))
    num_qubits = len(ordered_qubits)
    qubit_map = {qubit: index for index, qubit in enumerate(ordered_qubits)}

    opsums_and_qubit_counts = []
    for psum in psumlist:
      opsum = []
      opsum_qubits = set()
      for pstr in psum:
        opstring = qsim.OpString()
        opstring.weight = pstr.coefficient
        for q, pauli in pstr.items():
          op = pauli.on(q)
          opsum_qubits.add(q)
          qsimc.add_op_to_opstring(op, qubit_map, opstring)
        opsum.append(opstring)
      opsums_and_qubit_counts.append((opsum, len(opsum_qubits)))

    if initial_state is None:
      initial_state = 0
    if not isinstance(initial_state, (int, np.ndarray)):
      raise TypeError('initial_state must be an int or state vector.')
    if not isinstance(program, qsimc.QSimCircuit):
      program = qsimc.QSimCircuit(program, device=program.device)

    options = {}
    options.update(self.qsim_options)

    param_resolvers = study.to_resolvers(params)
    if isinstance(initial_state, np.ndarray):
      if initial_state.dtype != np.complex64:
        raise TypeError(f'initial_state vector must have dtype np.complex64.')
      input_vector = initial_state.view(np.float32)
      if len(input_vector) != 2**num_qubits * 2:
        raise ValueError(f'initial_state vector size must match number of qubits.'
          f'Expected: {2**num_qubits * 2} Received: {len(input_vector)}')

    results = []
    if _needs_trajectories(program):
      translator_fn_name = 'translate_cirq_to_qtrajectory'
      ev_simulator_fn = qsim.qtrajectory_simulate_expectation_values
    else:
      translator_fn_name = 'translate_cirq_to_qsim'
      ev_simulator_fn = qsim.qsim_simulate_expectation_values

    for prs in param_resolvers:
      solved_circuit = protocols.resolve_parameters(program, prs)
      translator_fn = getattr(solved_circuit, translator_fn_name)
      options['c'] = translator_fn(qubit_order)
      options['s'] = self.get_seed()
      options['n'] = num_qubits

      if isinstance(initial_state, int):
        evs = ev_simulator_fn(options, opsums_and_qubit_counts, initial_state)
      elif isinstance(initial_state, np.ndarray):
        evs = ev_simulator_fn(options, opsums_and_qubit_counts, input_vector)
      results.append(evs)

    return results
